/*	$NetBSD$	*/

/*-
 * Copyright (c) 2013 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Steffen Nurpmeso.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef _TEXT_H_
#define	_TEXT_H_

/*
 * Multibyte and multi-codepoint ("graphem") string ("text") interface.
 *
 * Support levels (value of TEXT_SUPPORT_LEVEL):
 * 1:
 *	Offers a basic set of character and multibyte string functions.
 *	It is usable in all ISO C90, Addendum 1 compliant environments.
 *	It does not introduce any internal intelligence, but solely builds
 *	upon the native environment.
 *	It is thus not thread safe (due to use of the global LC_CTYPE locale).
 *	Also, multi-codepoint strings are not supported for real, with all the
 *	consequences that this implies for the interface.
 *
 *	Keywords: const, restrict, [inline]
 *	Types: ssize_t, size_t, [wchar_t], [mbstate_t]
 *	[Funcs: mbrtowc(), isw*(), tow(upper|lower)(), wctomb(), wcwidth()]
 *	[Headers: wchar.h, wctype.h]
 *
 * TEXT_SUPPORT_WUCS
 *	If this is defined then the native wide character environment uses
 *	UCS4 (ISO 10646) or UTF-32 (Unicode).
 *	This means that, whereas multi-codepoint strings are still not
 *	supported, the by-codepoint comparison is truly well-defined, we can
 *	provide internal toupper() and tolower(), and txtwidth() will return
 *	correct results for any given codepoint.
 *	No other changes beyond support level 1.
 *
 * TEXT_SUPPORT_XLOCALE:
 *	If defined, an additional *_l() mirror of the interface of support
 *	level 1 is available (the *_l() family works with the charset of the
 *	specified locale_t argument instead of that of the LC_CTYPE one).
 *	The *_l() series is thread safe from our point of view.
 *	No other changes beyond support level 1.
 *
 *	Types: locale_t
 *	[Funcs: _l() versions of support level 1 functions]
 *	Headers: locale.h (assumes xlocale.h)
 *
 * 2:
 *	The same as support level 1 (including extensions), but knows how to
 *	deal with internals of the native interface.
 *	Thus it's thread safe (even for the non _l() interface series).
 *	XXX It can offer the txtby*() series (at least partially).
 *
 * 3:
 *	Knows wether input is in UTF-8, and is capable to convert input to
 *	UTF-8 for classification purposes on-the-fly if not, thread-safe.
 *	Internally uses Unicode properties to classify characters.
 *	Is charset-centric as opposed to locale-centric, and introduces a "s"
 *	function series that takes a "charset-descriptor" argument.
 *	This support level does not need any native wide character environment
 *	to function.
 *
 * 4:
 *	Level 4 would extend level 3 by also doing the character set
 *	conversion, thus offering iconv(3) via (loadable) modules, working with
 *	UTF-8 from top to bottom.
 *	This level could thus be used as a foundation for a complete POSIX
 *	string environment; it could offer wrappers for the *w* family of
 *	functions to accomplish that.
 */

#define TEXT_SUPPORT_LEVEL	1

/*
 * I don't know of any wchar_t implementation that internally bases on
 * something else than UCS / Unicode.  Due to this, use a blacklist.
 */
#if 1
# define TEXT_SUPPORT_WUCS
#endif

#if defined __APPLE__
# define TEXT_SUPPORT_XLOCALE
#endif

/* Undocumented switch, mostly for testing; also: UnixWare */
#if defined __APPLE__ || defined __DragonFly__  || defined __FreeBSD__ || \
	defined __NetBSD__ || defined __OpenBSD__
# define TEXT_SUPPORT_8BIT_C_LOCALE
#endif

#include <sys/cdefs.h>
#include <sys/types.h>
#ifdef TEXT_SUPPORT_XLOCALE
# include <locale.h>
# include <xlocale.h>
#endif

#ifndef NULL
# include <stddef.h>
#endif

/* XXX Simply defining restrict and inline is a multiple bits too rude */
#if ! defined __STDC_VERSION__ || __STDC_VERSION__ + 0 < 199901L
# ifdef __GNUC__
#  ifndef restrict
#   define restrict	__restrict
#  endif
#  ifndef inline
#   define inline	__inline
#  endif
# else
#  ifndef restrict
#   define restrict
#  endif
#  ifndef inline
#   define inline
#  endif
# endif
#endif

__BEGIN_DECLS

#define TXT_BOUND_MAX	(4 * 6)	/* XXX Check max. graphem length UTF-8! */

enum txttype {
	TT_ALPHA	= 1<< 0,
	TT_ASCII	= 1<< 1,
	TT_BLANK	= 1<< 2,
	TT_CNTRL	= 1<< 3,
	TT_DIGIT	= 1<< 4,
	TT_ALNUM	= TT_ALPHA | TT_DIGIT,
	TT_GRAPH	= 1<< 5,
	TT_LOWER	= 1<< 6,
	TT_PRINT	= 1<< 7,
	TT_PUNCT	= 1<< 8,
	TT_SPACE	= 1<< 9,
	TT_UPPER	= 1<<10,
	TT_XDIGIT	= 1<<11
	/* XXX ideogram,phonogram,special; better: Unicode properties.
	 * XXX those need more thinking, 'would excess bit range.
	 * XXX Plan9 (9atom) uses runeclass(), runecompose(), (is|to)baserune()
	 * XXX i.e., ("de") normalization
	 * XXX that is layer 3+ then, anyway */
};

ssize_t		txtbound(char const *);
ssize_t		txtnbound(char const *, size_t);

ssize_t		txtllcat(char *restrict, size_t, char const *restrict, size_t,
			size_t *);
ssize_t		txtncat(char *restrict, char const *restrict, size_t, size_t *);
ssize_t		txtcat(char *restrict, char const *restrict);
ssize_t		txtlcat(char *restrict, char const *restrict, size_t, size_t *);

ssize_t		txtcmp(char const *, char const *);
ssize_t		txtncmp(char const *, char const *, size_t);
ssize_t		txtcasecmp(char const *, char const *);
ssize_t		txtncasecmp(char const *, char const *, size_t);

ssize_t		txtcpy(char *restrict, char const *restrict);
ssize_t		txtlcpy(char *restrict, char const *restrict, size_t, size_t *);
ssize_t		txtllcpy(char *restrict, size_t, char const *restrict, size_t,
			size_t *);
ssize_t		txtncpy(char *restrict, char const *restrict, size_t);

char *		txtdup(char const *);
char *		txtndup(char const *, size_t, size_t *);

ssize_t		txtlen(char const *, size_t *);
ssize_t		txtnlen(char const *, size_t, size_t *);

char const *	txtstr(char const *, char const *);
char const *	txtllstr(char const *, size_t, char const *, size_t);
char const *	txtcasestr(char const *, char const *);
char const *	txtllcasestr(char const *, size_t, char const *, size_t);

char const *	txtrstr(char const *, char const *);
char const *	txtllrstr(char const *, size_t, char const *, size_t);
char const *	txtrcasestr(char const *, char const *);
char const *	txtllrcasestr(char const *, size_t, char const *, size_t);

ssize_t		txtwidth(char const *, size_t *);
ssize_t		txtnwidth(char const *, size_t, size_t *);
ssize_t		txtswidth(char const *, size_t, size_t *);

ssize_t		istalnum(char const *);
ssize_t		istlalnum(char const *, size_t);
ssize_t		istalpha(char const *);
ssize_t		istlalpha(char const *, size_t);
#if 0 /* XXX istascii(): not portable, only SUPPORT_LEVEL>3, <=U+007F ? */
ssize_t		istascii(char const *);
ssize_t		istlascii(char const *, size_t);
#endif
ssize_t		istblank(char const *);
ssize_t		istlblank(char const *, size_t);
ssize_t		istcntrl(char const *);
ssize_t		istlcntrl(char const *, size_t);
ssize_t		istdigit(char const *);
ssize_t		istldigit(char const *, size_t);
ssize_t		istgraph(char const *);
ssize_t		istlgraph(char const *, size_t);
ssize_t		istlower(char const *);
ssize_t		istllower(char const *, size_t);
ssize_t		istprint(char const *);
ssize_t		istlprint(char const *, size_t);
ssize_t		istpunct(char const *);
ssize_t		istlpunct(char const *, size_t);
ssize_t		istspace(char const *);
ssize_t		istlspace(char const *, size_t);
ssize_t		istupper(char const *);
ssize_t		istlupper(char const *, size_t);
ssize_t		istxdigit(char const *);
ssize_t		istlxdigit(char const *, size_t);

ssize_t		totlower(char *, char const *);
ssize_t		totllower(char *, char const *, size_t);
ssize_t		totupper(char *, char const *);
ssize_t		totlupper(char *, char const *, size_t);

#ifdef TEXT_SUPPORT_XLOCALE

ssize_t		txtbound_l(char const *, locale_t);
ssize_t		txtnbound_l(char const *, size_t, locale_t);

ssize_t		txtlcat_l(char *restrict, char const *restrict, size_t,
			size_t *, locale_t);
ssize_t		txtllcat_l(char *restrict, size_t, char const *restrict, size_t,
			size_t *, locale_t);
ssize_t		txtncat_l(char *restrict, char const *restrict, size_t,
			size_t *, locale_t);

ssize_t		txtcmp_l(char const *, char const *, locale_t);
ssize_t		txtncmp_l(char const *, char const *, size_t, locale_t);
ssize_t		txtcasecmp_l(char const *, char const *, locale_t);
ssize_t		txtncasecmp_l(char const *, char const *, size_t, locale_t);

ssize_t		txtlcpy_l(char *restrict, char const *restrict, size_t,
			size_t *, locale_t);
ssize_t		txtllcpy_l(char *restrict, size_t, char const *restrict, size_t,
			size_t *, locale_t);
ssize_t		txtncpy_l(char *restrict, char const *restrict, size_t,
			locale_t);

char *		txtndup_l(char const *, size_t, size_t *, locale_t);

ssize_t		txtlen_l(char const *, size_t *, locale_t);
ssize_t		txtnlen_l(char const *, size_t, size_t *, locale_t);

char const *	txtstr_l(char const *, char const *, locale_t);
char const *	txtllstr_l(char const *, size_t, char const *, size_t,
			locale_t);
char const *	txtcasestr_l(char const *, char const *, locale_t);
char const *	txtllcasestr_l(char const *, size_t, char const *, size_t,
			locale_t);

char const *	txtrstr_l(char const *, char const *, locale_t);
char const *	txtllrstr_l(char const *, size_t, char const *, size_t,
			locale_t);
char const *	txtrcasestr_l(char const *, char const *, locale_t);
char const *	txtllrcasestr_l(char const *, size_t, char const *, size_t,
			locale_t);

ssize_t		txtwidth_l(char const *, size_t *, locale_t);
ssize_t		txtnwidth_l(char const *, size_t, size_t *, locale_t);
ssize_t		txtswidth_l(char const *, size_t, size_t *, locale_t);

ssize_t		istalnum_l(char const *, locale_t);
ssize_t		istlalnum_l(char const *, size_t, locale_t);
ssize_t		istalpha_l(char const *, locale_t);
ssize_t		istlalpha_l(char const *, size_t, locale_t);
# if 0 /* XXX istascii_l(): not portable, only SUPPORT_LEVEL>3, <=U+007F ? */
ssize_t		istascii_l(char const *, locale_t);
ssize_t		istlascii_l(char const *, size_t, locale_t);
# endif
ssize_t		istblank_l(char const *, locale_t);
ssize_t		istlblank_l(char const *, size_t, locale_t);
ssize_t		istcntrl_l(char const *, locale_t);
ssize_t		istlcntrl_l(char const *, size_t, locale_t);
ssize_t		istdigit_l(char const *, locale_t);
ssize_t		istldigit_l(char const *, size_t, locale_t);
ssize_t		istgraph_l(char const *, locale_t);
ssize_t		istlgraph_l(char const *, size_t, locale_t);
ssize_t		istlower_l(char const *, locale_t);
ssize_t		istllower_l(char const *, size_t, locale_t);
ssize_t		istprint_l(char const *, locale_t);
ssize_t		istlprint_l(char const *, size_t, locale_t);
ssize_t		istpunct_l(char const *, locale_t);
ssize_t		istlpunct_l(char const *, size_t, locale_t);
ssize_t		istspace_l(char const *, locale_t);
ssize_t		istlspace_l(char const *, size_t, locale_t);
ssize_t		istupper_l(char const *, locale_t);
ssize_t		istlupper_l(char const *, size_t, locale_t);
ssize_t		istxdigit_l(char const *, locale_t);
ssize_t		istlxdigit_l(char const *, size_t, locale_t);

ssize_t		totlower_l(char *, char const *, locale_t);
ssize_t		totllower_l(char *, char const *, size_t, locale_t);
ssize_t		totupper_l(char *, char const *, locale_t);
ssize_t		totlupper_l(char *, char const *, size_t, locale_t);

#endif

__END_DECLS
#endif /* _TEXT_H_ */
