/*
 * Written by Steffen Nurpmeso.
 * Public domain.
 */

#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "text.h"

struct tsdat {
	char const *s;
	size_t len;
};

struct tdat {
	int no;
	char const *s;
	size_t n;
	size_t len;
};

static struct tsdat const	_tsd[] = {
#define X(S)	{ S, sizeof(S) - 1 },
	X("")
	X("#")
	X("#1")		X("#12")	X("#123")	X("#12345")
	X("#123456")	X("#1234567")	X("#12345678")	X("#123456789")
	X("#1234567890") X("#1234567")	X("#12345678")	X("#123456789")
	/* With UTF-8 */
	X("#\303\244")
	X("#\303\244b")
	X("#\303\244bc")
	X("#\303\244bcd")
	X("#\303\244bcde")
	X("#\303\244bcdef")
	X("#\303\244bcdefg")
	X("#\303\244bcdefgh")
	X("#\303\244bcdefghi")
	X("#\303\244bcdefghij")
	X("#\303\244bcdefghijk")
	X("#\303\244bcdefghijkl")
	X("#\303\244bcdefghijklm")
	X("#\303\244bcdefghijklmn")
	X("#\303\244bcdefghijklmno")
	X("#\303\244bcdefghijklmn\303\266")
	X("#\303\244bcdefghijklmn\303\266p")
	X("#\303\244bcdefghijklmn\303\266pq")
	X("#\303\244bcdefghijklmn\303\266pqr")
	X("#\303\244bcdefghijklmn\303\266pqrs")
	X("#\303\244bcdefghijklmn\303\266pqrst")
	X("#\303\244bcdefghijklmn\303\266pqrstu")
#undef X
	{ NULL, 0 }
};

static struct tdat const	_td1_posix[] = {
	{ 0, "", 0, 0 },
	{ 1, "a", 1, 1 },
	{ 2, "A", 1, 1 },
	{ 3, "\t", 1, 1 },
#ifdef TEXT_SUPPORT_8BIT_C_LOCALE
	{ 4, "\303\244", 2, 2 },
	{ 5, "\370\244", 2, 2 },
	{ 6, "\341\240\200", 3, 3 },
	{ 7, "\361\240\177", 3, 3 },
#endif
	{ 8, "a\t", 2, 2 },
	{ 9, "abc\t", 4, 4 },
	{ 10, "gamma", 5, 5 },
#ifdef TEXT_SUPPORT_8BIT_C_LOCALE
	{ 11, "gamma\303\244", 6, 6 },
	{ 12, "gamma\303\244", 7, 7 },
#endif
	{ 13, NULL, 0, 0 }
},
			_td1_utf8[] = {
	{ 0, "", 0, 0 },
	{ 1, "a", 1, 1 },
	{ 2, "A", 1, 1 },
	{ 3, "\t", 1, 1 },
	{ 4, "\303\244", 2, 2 },
	{ 5, "\370\244", 2, 0 },
	{ 6, "\341\240\200", 3, 3 },
	{ 7, "\361\240\200", 3, 0 },
	{ 8, "a\t", 2, 2 },
	{ 9, "abc\t", 4, 4 },
	{ 10, "gamma", 5, 5 },
	{ 11, "gamma\303\244", 6, 5 },
	{ 12, "gamma\303\244", 7, 7 },
	{ 0, NULL, 0, 0 }
};

static char const *	_ts;
static char *	(*	_tptf)(char const *, size_t, size_t *);
#ifdef TEXT_SUPPORT_XLOCALE
static locale_t		_lc;
#endif

static void	_t0(int tno);
static void	_t1(int tno, struct tdat const *);

static void
_t0(int tno)
{
	struct tsdat const *x;
	char *d;
	size_t r;

	for (x = _tsd; x->s != NULL; ++x) {
		d = txtdup(x->s);
		r = strlen(d);
		if (r != x->len)
			atf_tc_fail("%d txtdup(): got %zu, should be %zu\n",
				tno, r, x->len);
		if (memcmp(d, x->s, r + 1) != 0)
			atf_tc_fail("%d txtdup(): copy failure\n", tno);
		free(d);
	}
}

static void
_t1(int tno, struct tdat const *td)
{
	struct tdat const *x;
	char *d;
	size_t r, len;

	for (x = td; x->s != NULL; ++x) {
		d = (*_tptf)(x->s, x->n, &len);
		r = strlen(d);
		if (r != x->len)
			atf_tc_fail("%d.%d %s(): got %zu, should be %zu\n",
				tno, x->no, _ts, r, x->len);
		if (len != x->len)
			atf_tc_fail("%d.%d %s(): len is %zu, should be %zu\n",
				tno, x->no, _ts, len, x->len);
		if (r > 0 && memcmp(d, x->s, r) != 0)
			atf_tc_fail("%d.%d: %s(): copy failure\n",
				tno, x->no, _ts);
		if (d[r] != '\0')
			atf_tc_fail("%d.%d: %s(): termination failure\n",
				tno, x->no, _ts);
		free(d);
	}
}

ATF_TC_BODY(txtdup_go, tc)
{
	_ts = "POSIX:txtdup";
	if (setlocale(LC_CTYPE, "POSIX") == 0)
		atf_tc_fail("1. %s: cannot set locale\n", _ts);
	_t0(1);

	_ts = "en_USSR.UTF-8:txtdup";
	if (setlocale(LC_CTYPE, "en_US.UTF-8") == 0)
		atf_tc_fail("1. %s: cannot set locale\n", _ts);
	_t0(2);
}

static char *
_tgo(char const *s, size_t n, size_t *p)
{ (void)n; return txtndup(s, n, p); }

ATF_TC_BODY(txtndup_go, tc)
{
	_tptf = &_tgo;

	_ts = "POSIX:txtndup";
	if (setlocale(LC_CTYPE, "POSIX") == 0)
		atf_tc_fail("1. %s: cannot set locale\n", _ts);
	_t1(3, _td1_posix);

	_ts = "en_USSR.UTF-8:txtndup";
	if (setlocale(LC_CTYPE, "en_US.UTF-8") == 0)
		atf_tc_fail("1. %s: cannot set locale\n", _ts);
	_t1(4, _td1_utf8);
}

#ifdef TEXT_SUPPORT_XLOCALE
static char *
_tgo_l(char const *s, size_t n, size_t *p)
{ (void)n; return txtndup_l(s, n, p, _lc); }

ATF_TC_BODY(txtndup_l_go, tc)
{
	_tptf = &_tgo_l;

	_ts = "POSIX:txtndup_l";
	if ((_lc = newlocale(LC_CTYPE_MASK, "POSIX", (locale_t)0)) ==
			(locale_t)0)
		atf_tc_fail("2. %s: cannot newlocale()\n", _ts);
	_t1(5, _td1_posix);

	_ts = "en_USSR.UTF-8:txtndup_l";
	if ((_lc = newlocale(LC_CTYPE_MASK, "en_US.UTF-8", (locale_t)0)) ==
			(locale_t)0)
		atf_tc_fail("2. %s: cannot newlocale()\n", _ts);
	_t1(6, _td1_utf8);
}
#endif /* TEXT_SUPPORT_XLOCALE */
