/*
 *@ Note: this creates a test table using the wide environment of the running
 *@ system..
 *
 * Written by Steffen Nurpmeso.
 * Public domain.
 */

#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <wchar.h>
#include <wctype.h>

#include "text.h"

#if TEXT_SUPPORT_LEVEL > 2
# error t_txttrans builds a test table by using the local wide character
# error environment.  This results in a non-portable and incorrect test
# error table on some systems.
# error For TEXT_SUPPORT_LEVEL>2 we need a Unicode compliant table in here.
#endif

/* Note: we (yet) handle max. 2-byte UTF-8 sequences (< 0x800) */
#define COUNT	0x124

struct tdat {
	enum txttype flags;
	char utf8[16];
	char upper[16];
	char lower[16];
};

static struct tdat	_td[COUNT];

static char const *	_ts;
#ifdef TEXT_SUPPORT_XLOCALE
static locale_t		_lc;
#endif

static void	_create_table(void);
static void	_tgo(struct tdat const *);
#ifdef TEXT_SUPPORT_XLOCALE
static void	_tgo_l(struct tdat const *);
#endif

static void
_create_table(void)
{
	wchar_t i;

	if (setlocale(LC_CTYPE, "en_US.UTF-8") == 0)
		atf_tc_fail("0.1 Cannot setlocale() en_US.UTF-8 for table "
			"creation");

	for (i = 0; i < COUNT; ++i) {
		struct tdat *td = _td + i;
		td->flags = 0;
		if (iswalpha(i))
			td->flags |= TT_ALPHA;
		if (iswlower(i))
			td->flags |= TT_LOWER;
		if (iswupper(i))
			td->flags |= TT_UPPER;
		memset(td->utf8, 0, sizeof td->utf8);
		memset(td->upper, 0, sizeof td->upper);
		memset(td->lower, 0, sizeof td->lower);
		if (wctomb(td->utf8, i) < 0)
			atf_tc_fail("0.2 Cannot wctomb(3) for table creation");
		if (wctomb(td->upper, towupper(i)) < 0)
			atf_tc_fail("0.3 Cannot wctomb(3) for table creation");
		if (wctomb(td->lower, towlower(i)) < 0)
			atf_tc_fail("0.4 Cannot wctomb(3) for table creation");
	}
}

static void
_tgo(struct tdat const *td)
{
	char buf[TXT_BOUND_MAX * sizeof(_td[0].utf8)];
	size_t i;

	for (i = 0; i < COUNT; ++i, ++td) {
		ssize_t l = strlen(td->utf8), r, x;

		r = totlower(buf, td->utf8);
		if ((x = txtcmp(buf, td->lower)) != 0 || r < 0)
			atf_tc_fail("1.%zu.1 %s(): %zd; cmp: %zd\n",
				i, _ts, r, x);
		r = totllower(buf, td->utf8, l);
		if ((x = txtcmp(buf, td->lower)) != 0 || r < 0)
			atf_tc_fail("1.%zu.2 %s(): %zd; cmp: %zd\n",
				i, _ts, r, x);
		if (l > 1) {
			r = totllower(buf, td->utf8, l - 1);
			if (r >= 0)
				atf_tc_fail("1.%zu.3 %s(): %zd\n", i, _ts, r);
		}

		r = totupper(buf, td->utf8);
		if ((x = txtcmp(buf, td->upper)) != 0 || r < 0)
			atf_tc_fail("1.%zu.4 %s(): %zd; cmp: %zd\n",
				i, _ts, r, x);
		r = totlupper(buf, td->utf8, l);
		if ((x = txtcmp(buf, td->upper)) != 0 || r < 0)
			atf_tc_fail("1.%zu.5 %s(): %zd; cmp: %zd\n",
				i, _ts, r, x);
		if (l > 1) {
			r = totlupper(buf, td->utf8, l - 1);
			if (r >= 0)
				atf_tc_fail("1.%zu.6 %s(): %zd\n", i, _ts, r);
		}
	}
}

ATF_TC_BODY(txttrans_go, tc)
{
	_ts = "en_USSR.UTF-8:toXY()";
	if (setlocale(LC_CTYPE, "en_US.UTF-8") == 0)
		atf_tc_fail("1. %s: cannot set locale\n", _ts);
	_tgo(_td);
}

#ifdef TEXT_SUPPORT_XLOCALE
static void
_tgo_l(struct tdat const *td)
{
	char buf[TXT_BOUND_MAX * sizeof(_td[0].utf8)];
	size_t i;

	for (i = 0; i < COUNT; ++i, ++td) {
		ssize_t l = strlen(td->utf8), r, x;

		r = totlower_l(buf, td->utf8, _lc);
		if ((x = txtcmp(buf, td->lower)) != 0 || r < 0)
			atf_tc_fail("2.%zu.1 %s(): %zd; cmp: %zd\n",
				i, _ts, r, x);
		r = totllower_l(buf, td->utf8, l, _lc);
		if ((x = txtcmp(buf, td->lower)) != 0 || r < 0)
			atf_tc_fail("2.%zu.2 %s(): %zd; cmp: %zd\n",
				i, _ts, r, x);
		if (l > 1) {
			r = totllower_l(buf, td->utf8, l - 1, _lc);
			if (r >= 0)
				atf_tc_fail("2.%zu.3 %s(): %zd\n", i, _ts, r);
		}

		r = totupper_l(buf, td->utf8, _lc);
		if ((x = txtcmp(buf, td->upper)) != 0 || r < 0)
			atf_tc_fail("2.%zu.4 %s(): %zd; cmp: %zd\n",
				i, _ts, r, x);
		r = totlupper_l(buf, td->utf8, l, _lc);
		if ((x = txtcmp(buf, td->upper)) != 0 || r < 0)
			atf_tc_fail("2.%zu.5 %s(): %zd; cmp: %zd\n",
				i, _ts, r, x);
		if (l > 1) {
			r = totlupper_l(buf, td->utf8, l - 1, _lc);
			if (r >= 0)
				atf_tc_fail("2.%zu.6 %s(): %zd\n", i, _ts, r);
		}
	}
}

ATF_TC_BODY(txttrans_l_go, tc)
{
	_ts = "en_USSR.UTF-8:toXY_l()";
	if ((_lc = newlocale(LC_CTYPE_MASK, "en_US.UTF-8", (locale_t)0)) ==
			(locale_t)0)
		atf_tc_fail("2. %s: cannot newlocale()\n", _ts);
	_tgo_l(_td);
}
#endif /* TEXT_SUPPORT_XLOCALE */
