/*
 *@ Note: this creates a test table using the wide environment of the running
 *@ system..
 *
 * Written by Steffen Nurpmeso.
 * Public domain.
 */

#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <wchar.h>
#include <wctype.h>

#include "text.h"

#if TEXT_SUPPORT_LEVEL > 2
# error t_txttype builds a test table by using the local wide character
# error environment.  This results in a non-portable and incorrect test
# error table on some systems.
# error For TEXT_SUPPORT_LEVEL>2 we need a Unicode compliant table in here.
#endif

/* Note: we (yet) handle max. 2-byte UTF-8 sequences (< 0x800) */
#define COUNT	0x124

struct tdat {
	char utf8[8];
	enum txttype flags;
};

static struct tdat	_td[COUNT];

static char const *	_ts;
#ifdef TEXT_SUPPORT_XLOCALE
static locale_t		_lc;
#endif

static void	_create_table(void);
static void	_tgo(struct tdat const *);
#ifdef TEXT_SUPPORT_XLOCALE
static void	_tgo_l(struct tdat const *);
#endif

static void
_create_table(void)
{
	wchar_t i;

	if (setlocale(LC_CTYPE, "en_US.UTF-8") == 0)
		atf_tc_fail("0. Cannot setlocale() en_US.UTF-8 for table "
			"creation");

	for (i = 0; i < COUNT; ++i) {
		struct tdat *td = _td + i;
		memset(td->utf8, 0, sizeof td->utf8);
		if (i >= 0x80) {
			td->utf8[1] = (char)(0x80 | (i & 0x3F));
			td->utf8[0] = (char)(0xC0 | (i >> 6));
		} else
			td->utf8[0] = (char)i;
		td->flags = 0;
		if (iswalpha(i))
			td->flags |= TT_ALPHA;
		if (i < 0x80)
			td->flags |= TT_ASCII;
		if (iswblank(i))
			td->flags |= TT_BLANK;
		if (iswcntrl(i))
			td->flags |= TT_CNTRL;
		if (iswdigit(i))
			td->flags |= TT_DIGIT;
		/*if (iswalnum(i))
			td->flags |= TT_ALNUM; XXX implied: _ALPHA|_ALNUM */
		if (iswgraph(i))
			td->flags |= TT_GRAPH;
		if (iswlower(i))
			td->flags |= TT_LOWER;
		if (iswprint(i))
			td->flags |= TT_PRINT;
		if (iswpunct(i))
			td->flags |= TT_PUNCT;
		if (iswspace(i))
			td->flags |= TT_SPACE;
		if (iswupper(i))
			td->flags |= TT_UPPER;
		if (iswxdigit(i))
			td->flags |= TT_XDIGIT;
	}
}

static void
_tgo(struct tdat const *td)
{
	size_t i;

	for (i = 0; i < COUNT; ++i, ++td) {
		ssize_t l = strlen(td->utf8), x, r;

		x = (td->flags & TT_ALPHA) ? l : 0;
		r = istalpha(td->utf8);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_ALPHA, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istlalpha(td->utf8, l);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_ALPHA, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istlalpha(td->utf8, l - 1);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_ALPHA, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}
 
#if 0 /* XXX istascii(): not portable, only SUPPORT_LEVEL>3, <=U+007F ? */
		x = (td->flags & TT_ASCII) ? l : 0;
		r = istascii(td->utf8);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_ASCII, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istlascii(td->utf8, l);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_ASCII, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istlascii(td->utf8, l - 1);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_ASCII, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}
#endif

		x = (td->flags & TT_BLANK) ? l : 0;
		r = istblank(td->utf8);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_BLANK, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istlblank(td->utf8, l);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_BLANK, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istlblank(td->utf8, l - 1);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_BLANK, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}

		x = (td->flags & TT_CNTRL) ? l : 0;
		r = istcntrl(td->utf8);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_CNTRL, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istlcntrl(td->utf8, l);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_CNTRL, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istlcntrl(td->utf8, l - 1);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_CNTRL, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}

		x = (td->flags & TT_DIGIT) ? l : 0;
		r = istdigit(td->utf8);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_DIGIT, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istldigit(td->utf8, l);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_DIGIT, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istldigit(td->utf8, l - 1);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_DIGIT, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}

		x = (td->flags & TT_ALNUM) ? l : 0;
		r = istalnum(td->utf8);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_ALNUM, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istlalnum(td->utf8, l);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_ALNUM, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istlalnum(td->utf8, l - 1);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_ALNUM, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}

		x = (td->flags & TT_GRAPH) ? l : 0;
		r = istgraph(td->utf8);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_GRAPH, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istlgraph(td->utf8, l);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_GRAPH, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istlgraph(td->utf8, l - 1);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_GRAPH, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}

		x = (td->flags & TT_LOWER) ? l : 0;
		r = istlower(td->utf8);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_LOWER, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istllower(td->utf8, l);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_LOWER, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istllower(td->utf8, l - 1);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_LOWER, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}

		x = (td->flags & TT_PRINT) ? l : 0;
		r = istprint(td->utf8);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_PRINT, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istlprint(td->utf8, l);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_PRINT, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istlprint(td->utf8, l - 1);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_PRINT, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}

		x = (td->flags & TT_PUNCT) ? l : 0;
		r = istpunct(td->utf8);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_PUNCT, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istlpunct(td->utf8, l);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_PUNCT, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istlpunct(td->utf8, l - 1);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_PUNCT, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}

		x = (td->flags & TT_SPACE) ? l : 0;
		r = istspace(td->utf8);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_SPACE, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istlspace(td->utf8, l);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_SPACE, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istlspace(td->utf8, l - 1);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_SPACE, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}

		x = (td->flags & TT_UPPER) ? l : 0;
		r = istupper(td->utf8);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_UPPER, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istlupper(td->utf8, l);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_UPPER, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istlupper(td->utf8, l - 1);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_UPPER, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}

		x = (td->flags & TT_XDIGIT) ? l : 0;
		r = istxdigit(td->utf8);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_XDIGIT, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istlxdigit(td->utf8, l);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_XDIGIT, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istlxdigit(td->utf8, l - 1);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_XDIGIT, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}
	}
}

ATF_TC_BODY(txttype_go, tc)
{
	_ts = "en_USSR.UTF-8:istl?XY";
	if (setlocale(LC_CTYPE, "en_US.UTF-8") == 0)
		atf_tc_fail("1. %s: cannot set locale\n", _ts);
	_tgo(_td);
}

#ifdef TEXT_SUPPORT_XLOCALE
static void
_tgo_l(struct tdat const *td)
{
	size_t i;

	for (i = 0; i < COUNT; ++i, ++td) {
		ssize_t l = strlen(td->utf8), x, r;

		x = (td->flags & TT_ALPHA) ? l : 0;
		r = istalpha_l(td->utf8, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_ALPHA, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istlalpha_l(td->utf8, l, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_ALPHA, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istlalpha_l(td->utf8, l - 1, _lc);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_ALPHA, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}

#if 0 /* XXX istascii_l(): not portable, only SUPPORT_LEVEL>3, <=U+007F ? */
		x = (td->flags & TT_ASCII) ? l : 0;
		r = istascii_l(td->utf8, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_ASCII, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istlascii_l(td->utf8, l, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_ASCII, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istlascii_l(td->utf8, l - 1, _lc);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_ASCII, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}
#endif

		x = (td->flags & TT_BLANK) ? l : 0;
		r = istblank_l(td->utf8, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_BLANK, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istlblank_l(td->utf8, l, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_BLANK, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istlblank_l(td->utf8, l - 1, _lc);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_BLANK, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}

		x = (td->flags & TT_CNTRL) ? l : 0;
		r = istcntrl_l(td->utf8, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_CNTRL, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istlcntrl_l(td->utf8, l, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_CNTRL, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istlcntrl_l(td->utf8, l - 1, _lc);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_CNTRL, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}

		x = (td->flags & TT_DIGIT) ? l : 0;
		r = istdigit_l(td->utf8, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_DIGIT, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istldigit_l(td->utf8, l, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_DIGIT, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istldigit_l(td->utf8, l - 1, _lc);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_DIGIT, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}

		x = (td->flags & TT_ALNUM) ? l : 0;
		r = istalnum_l(td->utf8, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_ALNUM, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istlalnum_l(td->utf8, l, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_ALNUM, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istlalnum_l(td->utf8, l - 1, _lc);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_ALNUM, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}

		x = (td->flags & TT_GRAPH) ? l : 0;
		r = istgraph_l(td->utf8, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_GRAPH, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istlgraph_l(td->utf8, l, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_GRAPH, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istlgraph_l(td->utf8, l - 1, _lc);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_GRAPH, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}

		x = (td->flags & TT_LOWER) ? l : 0;
		r = istlower_l(td->utf8, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_LOWER, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istllower_l(td->utf8, l, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_LOWER, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istllower_l(td->utf8, l - 1, _lc);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_LOWER, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}

		x = (td->flags & TT_PRINT) ? l : 0;
		r = istprint_l(td->utf8, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_PRINT, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istlprint_l(td->utf8, l, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_PRINT, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istlprint_l(td->utf8, l - 1, _lc);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_PRINT, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}

		x = (td->flags & TT_PUNCT) ? l : 0;
		r = istpunct_l(td->utf8, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_PUNCT, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istlpunct_l(td->utf8, l, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_PUNCT, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istlpunct_l(td->utf8, l - 1, _lc);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_PUNCT, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}

		x = (td->flags & TT_SPACE) ? l : 0;
		r = istspace_l(td->utf8, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_SPACE, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istlspace_l(td->utf8, l, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_SPACE, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istlspace_l(td->utf8, l - 1, _lc);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_SPACE, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}

		x = (td->flags & TT_UPPER) ? l : 0;
		r = istupper_l(td->utf8, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_UPPER, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istlupper_l(td->utf8, l, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_UPPER, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istlupper_l(td->utf8, l - 1, _lc);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_UPPER, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}

		x = (td->flags & TT_XDIGIT) ? l : 0;
		r = istxdigit_l(td->utf8, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_XDIGIT, I: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		r = istlxdigit_l(td->utf8, l, _lc);
		if (r != x)
			atf_tc_fail("1.%zu %s(): TT_XDIGIT, II: "
				"got %zd, should be %zd\n", i, _ts, r, x);
		if (l > 1) {
			r = istlxdigit_l(td->utf8, l - 1, _lc);
			if (r >= 0)
				atf_tc_fail("1.%zu %s(): TT_XDIGIT, III: "
					"got %zd, should be <0\n", i, _ts, r);
		}
	}
}

ATF_TC_BODY(txttype_l_go, tc)
{
	_ts = "en_USSR.UTF-8:istl?XY_l";
	if ((_lc = newlocale(LC_CTYPE_MASK, "en_US.UTF-8", (locale_t)0)) ==
			(locale_t)0)
		atf_tc_fail("2. %s: cannot newlocale()\n", _ts);
	_tgo_l(_td);
}
#endif /* TEXT_SUPPORT_XLOCALE */
