/*
 * Written by Steffen Nurpmeso.
 * Public domain.
 */

#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "text.h"

struct tdat {
	int no;
	char const *s;
	size_t n;
	ssize_t ret;
	size_t passed;
};

static struct tdat const	_td1_posix[] = {
	{ 0, "", 0, 0, 0 },
	{ 1, "a", 1, 1, 1 },
	{ 2, "A", 1, 1, 1 },
	{ 3, "\t", 1, -1, 1 },
	/* NOTE: only if we don't rely on native wcwidth(3)!! */
#if defined TEXT_SUPPORT_8BIT_C_LOCALE && defined TEXT_SUPPORT_WUCS
	{ 4, "\303\244", 2, 1, 1 },
	{ 5, "\370\244", 2, 1, 1 },
	{ 6, "\341\240\200", 3, 1, 1 },
	{ 7, "\361\240\200", 3, 1, 1 },
#endif
	{ 0, NULL, 0, 0, 0 }
},
			_td1_utf8[] = {
	{ 0, "", 0, 0, 0 },
	{ 1, "a", 1, 1, 1 },
	{ 2, "A", 1, 1, 1 },
	{ 3, "\t", 1, -1, 1 },
	{ 4, "\303\244", 2, 1, 2 },
	{ 5, "\370\244", 2, -2, 0 },
#ifdef TEXT_SUPPORT_WUCS
	{ 6, "\341\240\200", 3, 1, 3 },
	{ 7, "\361\240\200", 3, -2, 0 },
#endif
	{ 0, NULL, 0, 0, 0 }
},
				_td2_posix[] = {
	{ 0, "", 0, 0, 0 },
	{ 1, "a", 1, 1, 1 },
	{ 2, "A", 1, 1, 1 },
	{ 3, "\t", 1, -1, 1 },
#if defined TEXT_SUPPORT_8BIT_C_LOCALE && defined TEXT_SUPPORT_WUCS
	{ 4, "\303\244", 2, 2, 2 },
	{ 5, "\370\244", 2, 2, 2 },
	{ 6, "\341\240\200", 3, -1, 3 },
	{ 7, "\361\240\177", 3, -1, 3 },
#endif
	{ 8, "a\t", 2, -1, 2 },
	{ 9, "abc\t", 4, -1, 4 },
	{ 10, "gamma", 5, 5, 5 },
#if defined TEXT_SUPPORT_8BIT_C_LOCALE && defined TEXT_SUPPORT_WUCS
	{ 11, "gamma\303\244", 6, 6, 6 },
	{ 12, "gamma\303\244", 7, 7, 7 },
#endif
	{ 13, NULL, 0, 0, 0 }
},
			_td2_utf8[] = {
	{ 0, "", 0, 0, 0 },
	{ 1, "a", 1, 1, 1 },
	{ 2, "A", 1, 1, 1 },
	{ 3, "\t", 1, -1, 1 },
	{ 4, "\303\244", 2, 1, 2 },
	{ 5, "\370\244", 2, -2, 0 },
	/* (UnixWare 7.1.4 native wcwidth(3) bails) */
#ifdef TEXT_SUPPORT_WUCS
	{ 6, "\341\240\200", 3, 1, 3 },
	{ 7, "\361\240\200", 3, -2, 0 },
#endif
	{ 8, "a\t", 2, -1, 2 },
	{ 9, "abc\t", 4, -1, 4 },
	{ 10, "gamma", 5, 5, 5 },
	{ 11, "gamma\303\244", 6, -7, 5 },
	{ 12, "gamma\303\244", 7, 6, 7 },
	{ 0, NULL, 0, 0, 0 }
};

static char const *	_ts;
static ssize_t	(*	_tptf)(char const *, size_t, size_t *);
#ifdef TEXT_SUPPORT_XLOCALE
static locale_t		_lc;
#endif

static void	_t1(int tno, struct tdat const *);

static void
_t1(int tno, struct tdat const *td)
{
	struct tdat const *x;
	ssize_t r;
	size_t passed;

	for (x = td; x->s != NULL; ++x) {
		r = (*_tptf)(x->s, x->n, &passed);
		if (r != x->ret)
			atf_tc_fail("%d.%d %s(): got %zd, should be %zd\n",
				tno, x->no, _ts, r, x->ret);
		if (passed != x->passed)
			atf_tc_fail("%d.%d %s(): passed is %zu, "
				"should be %zu\n",
				tno, x->no, _ts, passed, x->passed);
	}
}

static ssize_t
_tgo(char const *s, size_t n, size_t *p)
{ (void)n; return txtwidth(s, p); }

ATF_TC_BODY(txtwidth_go, tc)
{
	_tptf = &_tgo;

	_ts = "POSIX:txtwidth";
	if (setlocale(LC_CTYPE, "POSIX") == 0)
		atf_tc_fail("1. %s: cannot set locale\n", _ts);
	_t1(1, _td1_posix);

	_ts = "en_USSR.UTF-8:txtwidth";
	if (setlocale(LC_CTYPE, "en_US.UTF-8") == 0)
		atf_tc_fail("1. %s: cannot set locale\n", _ts);
	_t1(2, _td1_utf8);
}

#ifdef TEXT_SUPPORT_XLOCALE
static ssize_t
_tgo_l(char const *s, size_t n, size_t *p)
{ (void)n; return txtwidth_l(s, p, _lc); }

ATF_TC_BODY(txtwidth_l_go, tc)
{
	_tptf = &_tgo_l;

	_ts = "POSIX:txtwidth_l";
	if ((_lc = newlocale(LC_CTYPE_MASK, "POSIX", (locale_t)0)) ==
			(locale_t)0)
		atf_tc_fail("2. %s: cannot newlocale()\n", _ts);
	_t1(3, _td1_posix);

	_ts = "en_USSR.UTF-8:txtwidth_l";
	if ((_lc = newlocale(LC_CTYPE_MASK, "en_US.UTF-8", (locale_t)0)) ==
			(locale_t)0)
		atf_tc_fail("2. %s: cannot newlocale()\n", _ts);
	_t1(4, _td1_utf8);
}
#endif /* TEXT_SUPPORT_XLOCALE */

static ssize_t
_tngo(char const *s, size_t n, size_t *p)
{ return txtnwidth(s, n, p); }

ATF_TC_BODY(txtnwidth_go, tc)
{
	_tptf = &_tngo;

	_ts = "POSIX:txtnwidth";
	if (setlocale(LC_CTYPE, "POSIX") == 0)
		atf_tc_fail("3. %s: cannot set locale\n", _ts);
	_t1(5, _td1_posix);

	_ts = "en_USSR.UTF-8:txtnwidth";
	if (setlocale(LC_CTYPE, "en_US.UTF-8") == 0)
		atf_tc_fail("3. %s: cannot set locale\n", _ts);
	_t1(6, _td1_utf8);
}

#ifdef TEXT_SUPPORT_XLOCALE
static ssize_t
_tngo_l(char const *s, size_t n, size_t *p)
{ return txtnwidth_l(s, n, p, _lc); }

ATF_TC_BODY(txtnwidth_l_go, tc)
{
	_tptf = &_tngo_l;

	_ts = "POSIX:txtnwidth_l";
	if ((_lc = newlocale(LC_CTYPE_MASK, "POSIX", (locale_t)0)) ==
			(locale_t)0)
		atf_tc_fail("4. %s: cannot newlocale()\n", _ts);
	_t1(7, _td1_posix);

	_ts = "en_USSR.UTF-8:txtnwidth_l";
	if ((_lc = newlocale(LC_CTYPE_MASK, "en_US.UTF-8", (locale_t)0)) ==
			(locale_t)0)
		atf_tc_fail("4. %s: cannot newlocale()\n", _ts);
	_t1(8, _td1_utf8);
}
#endif /* TEXT_SUPPORT_XLOCALE */

static ssize_t
_tsgo(char const *s, size_t n, size_t *p)
{ return txtswidth(s, n, p); }

ATF_TC_BODY(txtswidth_go, tc)
{
	_tptf = &_tsgo;

	_ts = "POSIX:txtswidth";
	if (setlocale(LC_CTYPE, "POSIX") == 0)
		atf_tc_fail("3. %s: cannot set locale\n", _ts);
	_t1(9, _td2_posix);

	_ts = "en_USSR.UTF-8:txtswidth";
	if (setlocale(LC_CTYPE, "en_US.UTF-8") == 0)
		atf_tc_fail("3. %s: cannot set locale\n", _ts);
	_t1(10, _td2_utf8);
}

#ifdef TEXT_SUPPORT_XLOCALE
static ssize_t
_tsgo_l(char const *s, size_t n, size_t *p)
{ return txtswidth_l(s, n, p, _lc); }

ATF_TC_BODY(txtswidth_l_go, tc)
{
	_tptf = &_tsgo_l;

	_ts = "POSIX:txtswidth_l";
	if ((_lc = newlocale(LC_CTYPE_MASK, "POSIX", (locale_t)0)) ==
			(locale_t)0)
		atf_tc_fail("4. %s: cannot newlocale()\n", _ts);
	_t1(7, _td2_posix);

	_ts = "en_USSR.UTF-8:txtswidth_l";
	if ((_lc = newlocale(LC_CTYPE_MASK, "en_US.UTF-8", (locale_t)0)) ==
			(locale_t)0)
		atf_tc_fail("4. %s: cannot newlocale()\n", _ts);
	_t1(8, _td2_utf8);
}
#endif /* TEXT_SUPPORT_XLOCALE */
